---------------------------------------------------------------------------
-- This VHDL file was developed by Daniel Llamocca.  It may be
-- freely copied and/or distributed at no cost.  Any persons using this
-- file for any purpose do so at their own risk, and are responsible for
-- the results of such use.  Daniel Llamocca does not guarantee that
-- this file is complete, correct, or fit for any particular purpose.
-- NO WARRANTY OF ANY KIND IS EXPRESSED OR IMPLIED.  This notice must
-- accompany any copy of this file.
--------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- ***************************************************************************************************************
-- Vendor-specific comments for lpm_megafunctions:
-- ALTERA: The lpm megafunctions are embedded in the Quartus software.
--         'sqrt_nr' should always work. The parameter FAMILY is irrelevant (set it to Virtex-6)
-- XILINX: The 'fullypip' and 'combinatorial' cases always work.
--         The 'iterative' case only works well in V6/Spartan6 (Place & Route and Behavioral), not in Virtex-4
--            Virtex-4: It does not work well in V4 (Place & Route) (Behavioral is ok)
--            This is clearly a problem with the ISE synthesizer that uses different rules for V6 and V4 chips
--            In Place & Route: 'cin' is ignored, it always go to the default value
--                               Seems like the comparison with '-' goes wrong (problem with ISE synthesizer).
--         SOLUTION: for V4 devices, we have a generic 'FAMILY':
--            When set to 'Virtex-4', we use 'daniel_add_sub' as the adder/subractor in 'sqrt_nr_xe.vhd'.
-- ***************************************************************************************************************

-- *** USAGE *****************************************************************************************************
-- I/O Representation: unsigned fixed point:
-- Square root parameters:
--     N: number of bits of the operand (D). It has to be an even number
--     x: number of extra precision bits. It has to be an even number
--     NQ: number of bits of the square root (Q). NQ = N/2 + x

-- Algorithmically, when extra bits are needed, we have to add '2x' zeros to 'D'. This step is performed inside
-- the architecture, so the input 'D' will ALWAYS have 'N' bits.

-- If D is an integer number:
-- =========================
--   With N bits, the max. number is 2^N - 1.
--     The square root (integer, no fractional bits) would be: sqrt(2^N - 1) < 2^(N/2)
--       When dealing with integers, something lower than 2^(N/2) is lower or equal than 2^(N/2)-1:
--       ==> sqrt(2^N - 1) <= 2^(N/2) - 1
--       Therefore, the square root (integer) needs N/2 bits.
--   If x = 0, no point in Q
--   If x ~= 0, there is a point in Q, the extra 'x' bits are fractional bits.
--   Example:
--      D: d3d2d1d0 --> N = 4,  4 integer bits
--         If x = 0 (no extra bits) ==> Q = q1q0 --> NQ = 2
--         If x = 2 (2 extra bits) ==>  Q = q1q0.q-1q-2 -- > NQ = 4

-- If D is a fixed-point number:
-- ============================
--   With N bits, we can think of D as an integer multiplied by 2^(-nf), where nf: number of fractional bits.
--     The max. D then results:   (2^N - 1)*2^(-nf)
--     The square root would be:  sqrt(2^N-1)*sqrt(2^(-nf) < 2^(N/2)*2^(-nf/2)
--        Recall that sqrt(2^N-1) <= 2^(N/2) - 1
--        Thus, we can see that: sqrt(2^N-1)*sqrt(2^(-nf) <= (2^(N/2) - 1)*2^(-nf/2)
--     Therefore, the square root needs N/2 bits, but with (nf/2) fractional points.
--     As a result, the number of fractional bits (nf) should be even (as well as the number of integer bits)
--
--   So, 'x' would be extra fractional bits for more precision:
--   Example:
--      D: d3d2d1d0.d-1d-2d-3d-4 --> N = 8, 4 integer bits, 4 fractional bits
--         If x = 0 (no extra bits) ==> Q: q1q0.q-1q-2 --> NQ = 4
--         If x = 2 (2 extra bits)  ==> Q: q1q0.q-1q-2q-3q-4 --> NQ = 6
-- ****************************************************************************************************************

entity sqrt_nr is
	generic (N: INTEGER:= 12; -- must be even
				x: INTEGER:= 4; -- must be even
				IMPLEMENTATION: STRING:= "iterative"; -- "fullypip", "iterative", "combinatorial"
				FAMILY: STRING:= "Virtex-4"); -- Virtex-6, Virtex-4. When Virtex-4, sqrt_nr_xe uses 'daniel_add_sub'
	port (clock, resetn, s: in std_logic;
			D: in std_logic_vector (N-1 downto 0); -- operand
			Q: out std_logic_vector (N/2 + x - 1 downto 0); -- NQ = N/2 + x
			done: out std_logic);
end sqrt_nr;

architecture structure of sqrt_nr is

	constant NQ: INTEGER:= N/2 + x;

	component sqrt_nr_x_pip
		generic (N: INTEGER:= 12;
					x: INTEGER:= 0;
					NQ: INTEGER:= 6); -- NQ = N/2 + x
		port (clock, resetn,s: in std_logic;
				DI	: in std_logic_vector (N - 1 downto 0); -- operand
				QO	: out std_logic_vector (NQ - 1 downto 0); -- sqrt
				done: out std_logic);
	-- N = 2n, then NQ = n + x
	end component;
	
	component sqrt_nr_x_comb
		generic (N: INTEGER:= 12;
					x: INTEGER:= 8;
					NQ: INTEGER:= 14); -- NQ = N/2 + x
		port (clock, resetn, s: in std_logic;
				DI	: in std_logic_vector (N - 1 downto 0); -- operand
				QO	: out std_logic_vector (NQ - 1 downto 0); -- sqrt
				done: out std_logic);
	-- N = 2n, then NQ = n + x
	end component;
	
	component sqrt_nr_xe
		generic (N: INTEGER:= 22;
					x: INTEGER:= 4;
					NQ: INTEGER:= 15; -- NQ = N/2 + x;
					FAMILY: STRING:= "Virtex-6");  -- Virtex-6, Virtex-4. When Virtex-4, sqrt_nr_xe uses 'daniel_add_sub'
		port (clock, resetn, s: in std_logic;
				DI		: in std_logic_vector (N - 1 downto 0);
				QO		: out std_logic_vector (NQ - 1 downto 0);
				done	: out std_logic);
	end component;
	
begin

a0: assert (N rem 2 = 0)
	 report "N should be even!!!"
	 severity error;
	 
a1: assert (x rem 2 = 0)
	 report "x should be even!!!"
	 severity error;
	 
a2: assert (IMPLEMENTATION = "fullypip" or IMPLEMENTATION = "combinatorial" or IMPLEMENTATION = "iterative")
	 report "wrong IMPLEMENTATION parameter!"
	 severity error;
	 
pip: if IMPLEMENTATION = "fullypip" generate
			i1: sqrt_nr_x_pip generic map (N => N, x => x, NQ => NQ)
				 port map (clock => clock, resetn => resetn, s => s, DI => D, QO => Q, done => done);
	  end generate;

itr: if IMPLEMENTATION = "iterative" generate
			i2: sqrt_nr_xe generic map (N => N, x => x, NQ => NQ, FAMILY => FAMILY)
				 port map (clock => clock, resetn => resetn, s => s, DI => D, QO => Q, done => done);
	  end generate;
	 
cmb: if IMPLEMENTATION = "combinatorial" generate
			i3: sqrt_nr_x_comb generic map (N => N, x => x, NQ => NQ)
				 port map (clock => clock, resetn => resetn, s => s, DI => D, QO => Q, done => done);
	  end generate;
	 
end structure;
