---------------------------------------------------------------------------
-- This VHDL file was developed by Daniel Llamocca.  It may be
-- freely copied and/or distributed at no cost.  Any persons using this
-- file for any purpose do so at their own risk, and are responsible for
-- the results of such use.  Daniel Llamocca does not guarantee that
-- this file is complete, correct, or fit for any particular purpose.
-- NO WARRANTY OF ANY KIND IS EXPRESSED OR IMPLIED.  This notice must
-- accompany any copy of this file.
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

library work;
use work.lpm_components.all;

entity sqrt_nr_xe is
	generic (N: INTEGER:= 22;
				x: INTEGER:= 4;
				NQ: INTEGER:= 15; -- NQ = N/2 + x;
				FAMILY: STRING:= "Virtex-6");-- Virtex-6, Virtex-4. When Virtex-4, sqrt_nr_xe uses 'daniel_add_sub'
	port (clock, resetn, s: in std_logic;
			DI		: in std_logic_vector (N - 1 downto 0);
			QO		: out std_logic_vector (NQ - 1 downto 0);
			done	: out std_logic);
end sqrt_nr_xe;

architecture behavior of sqrt_nr_xe is

	function log_2(dato: in integer) return integer is
		variable i, valor: integer;
	begin
		i:= 0; valor:= dato;
		while valor /= 1 loop
			valor := valor - (valor/2); -- 'valor/2' performs truncation
			i:= i + 1;					-- Example: 15/2 = 7
		end loop;
		return i;
	end function log_2;
	
	component dffe
    port ( d : in  STD_LOGIC;
	        clrn: in std_logic:= '1';
			  prn: in std_logic:= '1';
           clk : in  STD_LOGIC;
			  ena: in std_logic;
           q : out  STD_LOGIC);
	end component;
	
	component daniel_ADD_SUB
		 generic ( lpm_width          : natural;    -- MUST be greater than 0
			        lpm_representation : string := "SIGNED";
			        lpm_direction      : string := "UNUSED"; -- UNUSED, ADD, SUB
			        cin_used				: string := "YES"); -- "YES", "NO"		 
		 port    ( dataa   : in std_logic_vector(lpm_width-1 downto 0);
			        datab   : in std_logic_vector(lpm_width-1 downto 0);
			        cin     : in std_logic := '0';		  
			        add_sub : in std_logic := '1'; -- 1: add, 0: sub
			        result   : out std_logic_vector(lpm_width-1 downto 0));
	end component;
	
	type state is (S1,S2);
	signal y: state;

	signal Q: std_logic_vector (NQ-1 downto 0);
	constant N_COUNT: integer:= log_2(NQ); -- integer log2 of NQ == # of bits of NQ
	signal L_ini, Eq, z, done_d: std_logic;
	signal R: std_logic_vector (NQ downto 0);
	signal Rp : std_logic_vector (NQ - 1 downto 0);
	signal Dp_even, Dp_odd, D_even, D_odd: std_logic_vector (NQ-x -1 downto 0);
	
	signal data_rR: std_logic_vector (NQ downto 0);
	signal datab_su: std_logic_vector (NQ-1 downto 0);
	signal cin_su: std_logic;

begin

	FSM_SQRT: process (resetn, clock)
	begin
		if resetn ='0' then
			y <= S1;
		elsif (clock'event and clock = '1') then
			case y is
				when S1 =>
					if s = '1' then y <= S2; else y <= S1; end if;
				when S2 =>
					if z = '1' then y <= S1; else y <= S2; end if;
			end case;
		end if;				
	end process;

	FSM_outputs: process (s, y, z)
	begin
		L_ini <= '0'; done_d <= '0'; Eq <='1';
		case y is
			when S1 =>
				if s = '0' then Eq <= '0'; end if; -- count in Q is stopped
				L_ini <= '1'; -- R <= 0; D <= initial value, cuenta <= maximum value
			when S2 =>
				if z = '1' then done_d <= '1'; end if;
		end case;
	end process;

dk: dffe port map (d => done_d, clrn => '1', prn => '1', clk => clock, ena => '1', q => done); 

-- Counter 'COUNT':
fn: lpm_counter generic map (LPM_WIDTH => N_COUNT, LPM_DIRECTION => "DOWN")
	 port map (data => conv_std_logic_vector(NQ-1, N_COUNT), clock => clock, sload => L_ini, cout => z);

-- Register Q:
rQ: lpm_shiftreg generic map (LPM_WIDTH => NQ, LPM_DIRECTION => "LEFT")
	 port map (clock => clock, enable => Eq, shiftin=> not(Rp(NQ-1)), sclr => L_ini , q => Q);

-- Register R:
rR: lpm_ff generic map (LPM_WIDTH => NQ + 1)
	 port map (data => data_rR, clock => clock, sclr => L_ini, q => R,
				  enable => '1', aclr => '0', aset => '0', aload => '0', sset => '0', sload =>'0');
	 data_rR <= Rp(NQ-2 downto 0)&(D_odd(NQ-x-1) xnor D_even(NQ-x-1))&not(D_even(NQ-x-1));

-- Register D => De: D_even (even bits), Do: D_odd (odd bits)
De: lpm_shiftreg generic map (LPM_WIDTH => NQ-x, LPM_DIRECTION => "LEFT")
	 port map (data => Dp_even, clock => clock, enable => '1', shiftin => '0', load => L_ini, q => D_even);
Do: lpm_shiftreg generic map (LPM_WIDTH => NQ-x, LPM_DIRECTION => "LEFT")
	 port map (data => Dp_odd, clock => clock, enable => '1', shiftin => '0', load => L_ini, q => D_odd);
	 
    -- DI is separated in even and odd bits. This is useful for shifting 2 bits ('00')
	 dis_DI: for i in 0 to NQ-x-1 generate
				   Dp_odd(i)  <= DI(2*i+1); -- odd bits
				   Dp_even(i) <= DI(2*i); -- even bits
			   end generate;

-- Adder with 'NQ' bits
V0: assert (family = "Virtex-4" or family = "Virtex-6")
	 report "Only Virtex-4 and Virtex-6 are allowed!"
	 severity error;

-- On Xilinx ISE synthesizer, 'lpm_add_sub' does not work properly for some devices (Virtex-4) when
-- the input 'cin' is used (it always go to default). As a result, a custom 'daniel_add_sub' function is used instead
V4: if family = "Virtex-4" generate
		sv: daniel_add_sub generic map (LPM_WIDTH => NQ, cin_used => "YES") 
			 port map (cin => cin_su, add_sub => R(NQ), dataa => R(NQ-1 downto 0), datab => datab_su, result => Rp);	
    end generate;
	 
V6: if family =  "Virtex-6" generate -- this also works for Spartan-6	 
		su: lpm_add_sub generic map (LPM_WIDTH => NQ) 
			 port map (cin => cin_su, add_sub => R(NQ), dataa => R(NQ-1 downto 0), datab => datab_su, result => Rp);	 
	 end generate;
	 
	 cin_su <= D_odd(NQ-x-1) or D_even(NQ-x-1);
	 datab_su <= '0'&Q(NQ-2 downto 0);

QO <= Q;				

end behavior;